#!/usr/bin/env python
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import cherrypy
import logging
import resource
import threading
from cmf.syssql.modules import ProcessesModule, ThreadsModule, ConnectionsModule, ZeroThroughNineModule, OneThroughTenThousandModule, HostsModule, MountsModule, InterfacesModule
from cmf.syssql.table import safe_access
try:
  import json
except:
  import simplejson as json

LOG = logging.getLogger(__name__)
LOG_FORMAT = '[%(asctime)s] %(module)-12s %(levelname)-8s %(message)s'
DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'
FORMATTER = logging.Formatter(LOG_FORMAT, DATE_FORMAT)

MODULES = [
  ProcessesModule,
  ThreadsModule,
  ConnectionsModule,
  HostsModule,
  MountsModule,
  InterfacesModule,

  # Debug Modules
  ZeroThroughNineModule,
  OneThroughTenThousandModule,

  # JVM modules are
  # JvmsModule,
  # JmxModule,
  # JstackModule,
]

class SysSQLServer(object):
  """
  Exposes /json_multi and /json URL to query syssql tables.
  """

  def __init__(self, agent, context_factory):
    self.agent = agent
    self.context_factory = context_factory
    self.log_handler = ThreadLocalLogHandler()
    self.log_handler.setLevel(logging.INFO)
    self.log_handler.setFormatter(FORMATTER)
    logging.getLogger().addHandler(self.log_handler)

  def __del__(self):
    logging.getLogger().removeHandler(self.log_handler)

  def json_multi(self, *args, **kwargs):
    """
    Retrieves multiple tables in one go.
    Expects ?tables=x&tables=y to retrieve two tables.
    """
    LOG.debug(kwargs)
    LOG.debug(args)
    try:
      # Capture logs for this request
      self.log_handler.enable()
      context = self.context_factory.create()
      results = []
      tables = kwargs.get("tables")
      if not isinstance(tables, list):
        # If only one table is specified, it's not
        # placed into a list.
        tables = [ tables ]
      for t in tables:
        chosen_module = context.by_name[t]
        if chosen_module is None:
          raise cherrypy.HTTPError(404)

        rows = chosen_module.fetch()

        context.update_resource_counters()
        table_results = dict(data=rows,
            table=chosen_module.name,
            logs=self.log_handler.fetch(),
            counters=context.counters)
        results.append(table_results)
        context.counters = {}
        self.log_handler.clear()

      return json.dumps(dict(tables=results), indent=2)
    finally:
      self.log_handler.disable()

class ContextFactory(object):
  def __init__(self, module_classes):
    self.module_classes = module_classes

  def create(self):
    modules = [ m() for m in self.module_classes ]
    return Context(modules)

CONTEXT_FACTORY = ContextFactory(MODULES)

class Context(object):
  """
  Captures module list and various counters.
  """
  def __init__(self, modules):
    self.counters = dict()
    for m in modules:
      m.set_context(self)
    self.modules = modules
    self.logs = []
    self.by_name = {}
    for m in self.modules:
      self.by_name[m.name] = m

    self.cpu_usage = self.get_cpu_usage()

  def get_cpu_usage(self):
    """
    Gets resource usage of this process with its children.  In a situation
    where there are concurrent things going on, this will over report, but
    it's a good proxy for how much CPU action is going on.
    """
    try:
      r1 = resource.getrusage(resource.RUSAGE_SELF)
      r2 = resource.getrusage(resource.RUSAGE_CHILDREN)
      return r1.ru_utime + r1.ru_stime + r2.ru_utime + r2.ru_stime
    except:
      return 0.0

  def increment(self, name, incr=1):
    """Increments a named counter."""
    cur = self.counters.get(name, 0)
    self.counters[name] = cur + incr

  def update_resource_counters(self):
    new_cpu_usage = self.get_cpu_usage()
    self.increment("cpu", new_cpu_usage - self.cpu_usage)
    self.cpu_usage = new_cpu_usage

class ThreadLocalLogHandler(logging.Handler):
  """
  Stores logs for the given thread.

  May be enabled and disabled.
  """
  def __init__(self):
    logging.Handler.__init__(self)
    self.buffers = threading.local()
    self.buffers.buf = None

  def enable(self):
    self.buffers.buf = []

  def clear(self):
    self.buffers.buf = []

  def disable(self):
    self.buffers.buf = None

  def fetch(self):
    return self.buffers.buf

  def emit(self, record):
    if not hasattr(self.buffers, "buf"):
      return

    if self.buffers.buf is not None:
      self.buffers.buf.append(self.format(record))
